/*
   ; K-TEAM debug calls library
   ; ==========================

   ;------------------------------------------------------------------------
   ; Author:    Jean Marc Koller 06/98
   ; Modifs:    Olivier Carmona  07/99
   ;
   ; Project:   KHEPERA
   ; Goal:              Source file of the K-TEAM Debug Library
   ;
   ;
   ;------------------------------------------------------------------------
 */

#include <string.h>
#include <sys/kos.h>
#define KT_INCLUDE
#include "ktdebug.h"

#define INFOTEXT	"\nKheperaTools debug monitor 0.0\n"

/* Maximum handler to be allocated. Max: 256 */
/* A higher MAXHANDLER uses more static memory! */
#define MAXHANDLER	32

/* Size of the event buffer. The larger the buffer,
 * the longer information can stay in the buffer (ring buffer) */
/* Warning: EVENTBUFFERSIZE should be a pow2 !! */
#define	EVENTBUFFERSIZE	(2*2*1024)

/* Buffer size to store variables for AutoWatch */
#define	WATCHBUFFER	1024

/******************* LOCAL VARIABLES *******************/

/* last allocated handler */
static unsigned char lastHandler;
/* maps the instrument invocated for each handler */
static unsigned char HandlerInstrument[MAXHANDLER];

/* ring buffer. Where imformation is saved in real-time (time-stamp+change) */
static unsigned char eventBuffer[EVENTBUFFERSIZE];
/* Current index in the eventbuffer */
static unsigned int eventBufferIndex;

#define	EMPTYINDEX	0xFFFFFFFF
/* Index in the eventbuffer for the remainings event to send */
static unsigned int tosendHandlerIndex[MAXHANDLER];
/* Bitmap of handlers to indicate which one is to send */
static volatile unsigned char tosendBitmap[MAXHANDLER / 8];
/* Length of variable to save for each handler */
static unsigned char tosendLength[MAXHANDLER];

/* Pointer to variable used by AutoWatch. */
/* Pointer to call-back used by Actor. */
static void *HandlerSpecificPtr[MAXHANDLER];
/* Nb variables to AutoWatch. Used to reduce processing time of */
/*  AutoWatch process if no variable to watch. */
static volatile unsigned char AutoWatchNbVar;
/* Bitmap of handlers that have requested AutoWatch. */
static volatile unsigned char AutoWatchBitmap[MAXHANDLER / 8];
/* Offset of each handler in AutoWatchBuffer. */
static unsigned char AutoWatchOffset[MAXHANDLER];
/* Buffer to store the last state of each variable. */
static unsigned char AutoWatchBuffer[WATCHBUFFER];

/* stores the last byte received by SER. */
static int sercarbuffer;

/*
#define	RESERVE_SER	while(ser_reserve_channel()==-1) tim_switch_fast();
#define	RELEASE_SER	ser_release_channel();
*/

static char informationText[] = INFOTEXT;
static const char textRestart[] = "restart\n";

/* INLINE append value to the event and check the pointer for end of buffer */
static inline void 
eventBufferAppend (int8 value)
{
  eventBuffer[eventBufferIndex++] = value;
  eventBufferIndex &= (EVENTBUFFERSIZE - 1);
}

/* INLINE set the corresponding bit in the bitmap of handler to send */
static inline void 
setTosendBitmap (unsigned char handler)
{
  tosendBitmap[handler >> 3] |= 1 << (handler & 7);
}

/* Send one handler related block to PC */
static void 
send (unsigned char handler)
{
  unsigned char buffer[64 + 12];
  unsigned char *wr = buffer;
  unsigned int index = tosendHandlerIndex[handler];
  unsigned char length = tosendLength[handler];
  *wr++ = '@';
  *wr++ = 'I';
  *wr++ = 'n';
  *wr++ = handler;
  while (length-- > 0)
    {
      *wr++ = eventBuffer[index++];	/* Copy variable into send buffer */
      index &= (EVENTBUFFERSIZE - 1);
    }
  *wr++ = 'K';
  RESERVE_SER
    ser_send_buffer (buffer, wr - buffer);
  RELEASE_SER
}

/* If information text has to be sent back (reception of '?' */
static volatile boolean sendInformation;

/* routine to send a zero-terminated text */
/* replaces the use of printf */
static void 
sendText (char *text)
{
  char *rd = text;
  int length = 0;
  while (*rd++)
    length++;
  RESERVE_SER
    ser_send_buffer (text, length);
  RELEASE_SER
}

/* The process which sends the debug information to the PC */
static void 
processSerialSend ()
{
  unsigned char hasFound;
  for (;;)
    {
      unsigned char bc;
      hasFound = 0;
      for (bc = 0; bc < MAXHANDLER / 8; bc++)
	{
	  unsigned char v = tosendBitmap[bc];
	  if (v)
	    {
	      unsigned char bb;
	      for (bb = 0; bb < 8; bb++, v >>= 1)
		{
		  if (v & 1)
		    {
		      unsigned char handler = 8 * bc + bb;
		      tosendBitmap[bc] &= ~(1 << bb);
		      send (handler);
		      hasFound = 1;
		    }
		}
	    }
	}
      if (sendInformation)
	{
	  sendText (informationText);
	  sendInformation = FALSE;
	}
#if 0
      if (!hasFound)		/* Nothing has been sended in this turn */
	tim_switch_fast ();	/* switch to other process */
#else
      tim_suspend_task (20);	/* Let the PC update the GUI */
#endif
    }
}

static void 
processSerialListen ()
{
  unsigned char indexRestart = 0;
  unsigned char iState = 0;
  unsigned char curHandler = 0, nbparams = 0, curparam = 0;
  unsigned char parBuffer[4];
  int car;
  for (;;)
    {
      tim_wait_sync (KSERByteReceived);
      car = ser_receive_byte ();
      sercarbuffer = car;
      if (car == -1)
	continue;
      if (car == textRestart[indexRestart])
	{
	  if (indexRestart == 7)
	    bios_restart_robot ();
	  else
	    indexRestart++;
	}
      else
	indexRestart = 0;
      if (car == '?')
	sendInformation = TRUE;
      switch (iState)
	{
	case 0:
	  iState = (car == '@') ? 1 : 0;
	  break;
	case 1:
	  iState = (car == 'I') ? 2 : 0;
	  break;
	case 2:
	  if (car == 'b')
	    iState = 3;
	  else if (car == 'f')
	    iState = 7;
	  else
	    iState = 0;
	  break;
	case 3:
	  curHandler = car;
	  iState++;
	  break;
	case 4:
	  nbparams = car;
	  curparam = 0;
	  iState = (nbparams > 0) ? 5 : 6;
	  break;
	case 5:
	  parBuffer[curparam++] = car;
	  if (curparam >= nbparams)
	    iState++;
	  break;
	case 6:
	  if (car == 'K')
	    {
	      void *ptr = HandlerSpecificPtr[curHandler];
	      if (ptr)
		{
		  if (nbparams == 0)
		    {
		      KTdebug_update (curHandler, NULL);
		      ((EventCallBackFunc) ptr) (0);
		    }
		  else if (nbparams == 4)
		    {
		      int value = (parBuffer[0] << 24) | (parBuffer[1] << 16) | (parBuffer[2] << 8) | parBuffer[3];
		      KTdebug_update (curHandler, &value);
		      ((EventCallBackFunc) ptr) (value);
		    }
		}
	    }
	  iState = 0;
	case 7:
	  iState = (car == 'R') ? 8 : 0;
	  break;
	case 8:
	  iState = (car == 'e') ? 9 : 0;
	  break;
	case 9:
	  iState = (car == 'E') ? 10 : 0;
	  break;
	case 10:
	  iState = (car == 'z') ? 11 : 0;
	  break;
	case 11:
	  iState = (car == 'e') ? 12 : 0;
	  break;
	case 12:
	  iState = (car == 'K') ? 13 : 0;
	  break;
	case 13:
	  iState = 0;
	  KTdebug_Freeze ();
	  break;
	default:
	  iState = 0;
	  break;
	}
    }
}

static void 
processAutoWatch ()
{
  for (;;)
    {
      unsigned char bc;
      while (AutoWatchNbVar == 0)
	tim_switch_fast ();
      for (bc = 0; bc < MAXHANDLER / 8; bc++)
	{
	  unsigned char v = AutoWatchBitmap[bc];
	  if (v)
	    {
	      unsigned char bb;
	      unsigned char handler = 8 * bc;
	      for (bb = 0; bb < 8; bb++, v >>= 1, handler++)
		{
		  if (v & 1)
		    {
		      unsigned char *refp = &AutoWatchBuffer[AutoWatchOffset[handler]];
		      unsigned char *varp = HandlerSpecificPtr[handler];
		      unsigned char length = tosendLength[handler];
		      if (memcmp (refp, varp, length) != 0)
			{
			  KTdebug_update (handler, HandlerSpecificPtr[handler]);
			  memcpy (refp, varp, length);
			}
		    }
		}
	    }
	}
      tim_switch_fast ();
    }
}

void 
KTdebug_init ()
{
  static const char senderTaskId[] = "KTdebug_sender";
  static const char ListenerTaskId[] = "KTdebug_listener";
  static const char AutoWatchTaskId[] = "KTdebug_autowatch";
  int count;
  
  ser_reset ();
  sercarbuffer = -1;
  lastHandler = 0;
  eventBufferIndex = 0;
  AutoWatchNbVar = 0;
  
  for (count = EVENTBUFFERSIZE; count > 0;)
    eventBuffer[--count] = '$';
    
  for (count = MAXHANDLER; count > 0;)
    {
      tosendHandlerIndex[--count] = 0;
      HandlerInstrument[count] = 0xFF;
      tosendLength[count] = 0;
      HandlerSpecificPtr[count] = NULL;
    }
    
  for (count = MAXHANDLER / 8; count > 0;)
    {
      tosendBitmap[--count] = 0;
      AutoWatchBitmap[count] = 0;
    }
    
  AutoWatchOffset[0] = 0;
  sendInformation = FALSE;
  tim_suspend_task (200);
  install_task ((char *) senderTaskId, 800, processSerialSend);
  install_task ((char *) ListenerTaskId, 800, processSerialListen);
  install_task ((char *) AutoWatchTaskId, 800, processAutoWatch);
}

#define	INSTR_BOOLEAN	0
#define	INSTR_INTEGER	1
#define	INSTR_ARRAY	2
#define	INSTR_BUTTON	3
#define	INSTR_VALUE	4

unsigned char 
KTdebug_addWatch_Boolean (char *name)
{
  unsigned char buffer[64 + 12];
  unsigned char *wr = buffer;
  unsigned char chandler;
  RESERVE_SER
    chandler = lastHandler++;
  *wr++ = '@';
  *wr++ = 'I';
  *wr++ = 'a';
  *wr++ = chandler;
  *wr++ = HandlerInstrument[chandler] = INSTR_BOOLEAN;
  *wr++ = 'J';
  while (*name)
    *wr++ = *name++;
  *wr++ = 0;
  *wr++ = 'M';
  *wr++ = 0;
  *wr++ = 'K';
  ser_send_buffer (buffer, wr - buffer);
  RELEASE_SER
    tosendLength[chandler] = 1;
  if (chandler + 1 < MAXHANDLER)
    AutoWatchOffset[chandler + 1] = AutoWatchOffset[chandler] + tosendLength[chandler];
  tim_suspend_task (100);
  return chandler;
}

unsigned char 
KTdebug_addWatch_Integer (char *name, enum IntegerModes mode, unsigned char size, unsigned char isUnsigned)
{
  unsigned char buffer[64 + 12];
  unsigned char *wr = buffer;
  unsigned char chandler;
  RESERVE_SER
    chandler = lastHandler++;
  if (size == 3)
    size = 4;
  else if (size > 4)
    size = 8;
  *wr++ = '@';
  *wr++ = 'I';
  *wr++ = 'a';
  *wr++ = chandler;
  *wr++ = HandlerInstrument[chandler] = INSTR_INTEGER;
  *wr++ = 'J';
  while (*name)
    *wr++ = *name++;
  *wr++ = 0;
  *wr++ = 'M';
  *wr++ = 1 + 1;
  *wr++ = mode;
  *wr++ = ((isUnsigned) ? 0 : 16) + size;
  *wr++ = 'K';
  ser_send_buffer (buffer, wr - buffer);
  RELEASE_SER
    tosendLength[chandler] = size;
  if (chandler + 1 < MAXHANDLER)
    AutoWatchOffset[chandler + 1] = AutoWatchOffset[chandler] + tosendLength[chandler];
  tim_suspend_task (100);
  return chandler;
}

#define WR32U(a)	*wr++=((int)a)>>24; *wr++=((int32)a)>>16; *wr++=((int32)a)>>8; *wr++=((int32)a);
unsigned char 
KTdebug_addWatch_IntegerMMO (char *name, enum IntegerModes mode, unsigned char size, unsigned char isUnsigned, int min, int max, int offset)
{
  unsigned char buffer[64 + 12];
  unsigned char *wr = buffer;
  unsigned char chandler;
  RESERVE_SER
    chandler = lastHandler++;
  if (size == 3)
    size = 4;
  else if (size > 4)
    size = 8;
  *wr++ = '@';
  *wr++ = 'I';
  *wr++ = 'a';
  *wr++ = chandler;
  *wr++ = HandlerInstrument[chandler] = INSTR_INTEGER;
  *wr++ = 'J';
  while (*name)
    *wr++ = *name++;
  *wr++ = 0;
  *wr++ = 'M';
  *wr++ = 1 + 1 + 4 + 4 + 4;
  *wr++ = mode;
  *wr++ = ((isUnsigned) ? 0 : 16) + size;
  WR32U (min);
  WR32U (max);
  WR32U (offset);
  *wr++ = 'K';
  ser_send_buffer (buffer, wr - buffer);
  RELEASE_SER
    tosendLength[chandler] = size;
  if (chandler + 1 < MAXHANDLER)
    AutoWatchOffset[chandler + 1] = AutoWatchOffset[chandler] + tosendLength[chandler];
  tim_suspend_task (100);
  return chandler;
}

unsigned char 
KTdebug_addWatch_Array (char *name, enum IntegerModes mode, unsigned char nbvars, struct VarArrayDesc varray[])
{
  unsigned char buffer[32];
  unsigned char *wr = buffer;
  unsigned char i;
  unsigned char length = 0;
  unsigned char chandler;
  RESERVE_SER
    chandler = lastHandler++;
  *wr++ = '@';
  *wr++ = 'I';
  *wr++ = 'a';
  *wr++ = chandler;
  *wr++ = HandlerInstrument[chandler] = INSTR_ARRAY;
  *wr++ = 'J';
  while (*name)
    *wr++ = *name++;
  *wr++ = 0;
  *wr++ = 'M';
  *wr++ = nbvars * 2 + 1 + 1;
  *wr++ = mode;
  *wr++ = nbvars;
  for (i = 0; i < nbvars; i++)
    {
      *wr++ = varray[i].size;
      *wr++ = varray[i].alpha;
      length += (varray[i].size >= 0) ? varray[i].size : -varray[i].size;
    }
  *wr++ = 'K';
  ser_send_buffer (buffer, wr - buffer);
  RELEASE_SER
    tosendLength[chandler] = length;
  if (chandler + 1 < MAXHANDLER)
    AutoWatchOffset[chandler + 1] = AutoWatchOffset[chandler] + tosendLength[chandler];
  tim_suspend_task (100);
  return chandler;
}

unsigned char 
KTdebug_addWatch_ArraySingleType (char *name, enum IntegerModes mode, unsigned char nbvars, char size, char *alpha, int min, int max)
{
  unsigned char buffer[64];
  unsigned char *wr = buffer;
  unsigned char i;
  unsigned char length = 0;
  unsigned char chandler;
  char *rdN;
  RESERVE_SER
    chandler = lastHandler++;
  *wr++ = '@';
  *wr++ = 'I';
  *wr++ = 'a';
  *wr++ = chandler;
  *wr++ = HandlerInstrument[chandler] = INSTR_ARRAY;
  *wr++ = 'J';
  while (*name)
    *wr++ = *name++;
  *wr++ = 0;
  *wr++ = 'M';
  *wr++ = nbvars * 2 + 1 + 1 + 2 * 4;
  *wr++ = mode | MODE_MINMAX;
  *wr++ = nbvars;
  WR32U (min);
  WR32U (max);
  rdN = alpha;
  for (i = 0; i < nbvars; i++)
    {
      *wr++ = size;
      *wr++ = (*rdN) ? *rdN++ : ' ';
      length += (size >= 0) ? size : -size;
    }
  *wr++ = 'K';
  ser_send_buffer (buffer, wr - buffer);
  RELEASE_SER
    tosendLength[chandler] = length;
  if (chandler + 1 < MAXHANDLER)
    AutoWatchOffset[chandler + 1] = AutoWatchOffset[chandler] + tosendLength[chandler];
  tim_suspend_task (100);
  return chandler;
}

unsigned char 
KTdebug_addActor_Button (char *name, EventCallBackFunc pCBfunc)
{
  unsigned char buffer[64 + 12];
  unsigned char *wr = buffer;
  unsigned char chandler;
  RESERVE_SER
    chandler = lastHandler++;
  *wr++ = '@';
  *wr++ = 'I';
  *wr++ = 'a';
  *wr++ = chandler;
  *wr++ = HandlerInstrument[chandler] = INSTR_BUTTON;
  *wr++ = 'J';
  while (*name)
    *wr++ = *name++;
  *wr++ = 0;
  *wr++ = 'M';
  *wr++ = 0;
  *wr++ = 'K';
  ser_send_buffer (buffer, wr - buffer);
  RELEASE_SER
    tosendLength[chandler] = 0;
  HandlerSpecificPtr[chandler] = pCBfunc;
  if (chandler + 1 < MAXHANDLER)
    AutoWatchOffset[chandler + 1] = AutoWatchOffset[chandler];
  tim_suspend_task (100);
  return chandler;
}

unsigned char 
KTdebug_addActor_Value (char *name, enum IntegerModes mode, int initial, int min, int max, EventCallBackFunc pCBfunc)
{
  unsigned char buffer[64 + 12];
  unsigned char *wr = buffer;
  unsigned char chandler;
  RESERVE_SER
    chandler = lastHandler++;
  *wr++ = '@';
  *wr++ = 'I';
  *wr++ = 'a';
  *wr++ = chandler;
  *wr++ = HandlerInstrument[chandler] = INSTR_VALUE;
  *wr++ = 'J';
  while (*name)
    *wr++ = *name++;
  *wr++ = 0;
  *wr++ = 'M';
  *wr++ = 1 + 4 + 2 * 4;
  *wr++ = mode;
  WR32U (initial);
  WR32U (min);
  WR32U (max);
  *wr++ = 'K';
  ser_send_buffer (buffer, wr - buffer);
  RELEASE_SER
    tosendLength[chandler] = 4;
  HandlerSpecificPtr[chandler] = pCBfunc;
  if (chandler + 1 < MAXHANDLER)
    AutoWatchOffset[chandler + 1] = AutoWatchOffset[chandler];
  tim_suspend_task (100);
  return chandler;
}

void 
KTdebug_setAutoWatch (unsigned char handler, void *ptr)
{
  HandlerSpecificPtr[handler] = ptr;
  memcpy (&AutoWatchBuffer[AutoWatchOffset[handler]], ptr, tosendLength[handler]);
  KTdebug_update (handler, ptr);
  AutoWatchNbVar--;
  AutoWatchBitmap[handler >> 3] |= 1 << (handler & 7);
}

void 
KTdebug_disAutoWatch (unsigned char handler)
{
  if (AutoWatchBitmap[handler >> 3] & (1 << (handler & 7)))
    return;
  AutoWatchBitmap[handler >> 3] &= ~(1 << (handler & 7));
  AutoWatchNbVar--;
}

void 
KTdebug_update (unsigned char handler, void *pvalue)
{
  unsigned int ctic = tim_get_ticcount ();
  unsigned char length = tosendLength[handler];
  unsigned char *rd = (unsigned char *) pvalue;
  eventBufferAppend ('@');
  eventBufferAppend (handler);
  eventBufferAppend (ctic >> 24);
  eventBufferAppend (ctic >> 16);
  eventBufferAppend (ctic >> 8);
  eventBufferAppend (ctic);
  tosendHandlerIndex[handler] = eventBufferIndex;
  for (; length > 0; length--)
    eventBufferAppend (*rd++);
  setTosendBitmap (handler);
}

static inline void 
serialWriteByte (unsigned char b)
{
  while (ser_tx_status ());
  ser_send_buffer (&b, 1);
}

static void 
sendEventBuffer ()
{
  unsigned int count = EVENTBUFFERSIZE, index;
  serialWriteByte ('@');
  serialWriteByte ('I');
  serialWriteByte ('x');
  serialWriteByte (EVENTBUFFERSIZE >> 24);
  serialWriteByte (EVENTBUFFERSIZE >> 16);
  serialWriteByte (EVENTBUFFERSIZE >> 8);
  serialWriteByte ((unsigned char) EVENTBUFFERSIZE);
  serialWriteByte ('J');
  index = eventBufferIndex;
  while (count-- > 0)
    {
      serialWriteByte (eventBuffer[index++]);
      index &= (EVENTBUFFERSIZE - 1);
    }
  serialWriteByte ('K');
}

void 
KTdebug_Freeze ()
{
  static const char sysF[] = "System frozen\n@IfrozenK";
  static const char sysActivity[] = "Resuming activity...\n";
  static const char thawStr[] = "@IthawK";
  static const char spitStr[] = "@IspitK";
  int car;
  unsigned char indexRestart = 0, indexThaw = 0, indexSpit = 0;
  
  tim_lock ();
  sendText ((char *) sysF);
  for (;;)
    {
      while (!ser_rx_status ());
      sercarbuffer = car = ser_receive_byte ();
      if (car == -1)
	continue;
      if (car == textRestart[indexRestart])
	{
	  if (indexRestart == 7)
	    bios_restart_robot ();
	  else
	    indexRestart++;
	}
      else
	indexRestart = 0;
      if (car == thawStr[indexThaw])
	{
	  if (indexThaw == 6)
	    break;
	  else
	    indexThaw++;
	}
      else
	indexThaw = 0;
      if (car == spitStr[indexSpit])
	{
	  if (indexSpit == 6)
	    sendEventBuffer ();
	  else
	    indexSpit++;
	}
      else
	indexSpit = 0;
      if (car == '?')
	sendText (informationText);
    }
  sendText ((char *) sysActivity);
  tim_unlock ();
}


int32 
KT_ser_receive_byte (void)
{
  int t = sercarbuffer;
  sercarbuffer = -1;
  return t;
}

int32 
KT_ser_rx_status (void)
{
  return (sercarbuffer == -1) ? 0 : -1;
}

#ifdef KHEPERA
static unsigned char sensorHandler;

static void 
spySensorsProcess ()
{
  unsigned short *ptr = (sens_get_pointer())->oProximitySensor;
  
  for (;;)
    {
      tim_wait_sync (1 << KSENSStartSweep);
      KTdebug_update (sensorHandler, (void *) ptr);
    }
}

void 
startSpySensors (enum IntegerModes mode)
{
  static const char SensorSpyTaskId[] = "KTdebug_listener";
  sensorHandler = KTdebug_addWatch_ArraySingleType ("Sens", mode, 8, -2, "01234567", 0, 1024);
  install_task ((char *) SensorSpyTaskId, 800, spySensorsProcess);
}
#endif /* KHEPERA */
